#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import sys
import traceback
import types

from fdk import context
from fdk import customer_code
from fdk import errors
from fdk import log
from fdk import response


async def with_deadline(ctx: context.InvokeContext,
                        handler_code: customer_code.Function,
                        data: io.BytesIO):
    """
    Runs function within a timer
    :param ctx: invoke context
    :type ctx: fdk.context.InvokeContext
    :param handler_code: customer's code
    :type handler_code: fdk.customer_code.Function
    :param data: request data stream
    :type data: io.BytesIO
    :return:
    """

    # ctx.Deadline() would never be an empty value,
    # by default it will be 30 secs from now

    try:
        handle_func = handler_code.handler()
        result = handle_func(ctx, data=data)
        if isinstance(result, types.CoroutineType):
            return await result

        return result
    except (Exception, TimeoutError) as ex:
        raise ex


async def handle_request(handler_code, format_def, **kwargs):
    """
    Handles a function's request
    :param handler_code: customer's code
    :type handler_code: fdk.customer_code.Function
    :param format_def: function's format
    :type format_def: str
    :param kwargs: request-specific parameters
    :type kwargs: dict
    :return: function's response
    :rtype: fdk.response.Response
    """
    log.log("in handle_request")
    ctx, body = context.context_from_format(format_def, **kwargs)
    log.set_request_id(ctx.CallID())
    log.log("context provisioned")
    try:
        response_data = await with_deadline(ctx, handler_code, body)
        log.log("function result obtained")
        if isinstance(response_data, response.Response):
            return response_data
        headers = ctx.GetResponseHeaders()
        log.log("response headers obtained")
        return response.Response(
            ctx, response_data=response_data,
            headers=headers, status_code=200)

    except TimeoutError as ex:
        log.log("Function timeout: {}".format(ex))
        (exctype, value, tb) = sys.exc_info()
        tb_flat = ''.join(
            s.replace('\n', '\\n') for s in traceback.format_tb(tb))
        log.get_request_log().error('{}:{}'.format(value, tb_flat))
        return errors.DispatchException(ctx, 504, str(ex)).response()
    except Exception as ex:
        log.log("exception appeared: {0}".format(ex))
        (exctype, value, tb) = sys.exc_info()
        tb_flat = ''.join(
            s.replace('\n', '\\n') for s in traceback.format_tb(tb))
        log.get_request_log().error('{}:{}'.format(value, tb_flat))
        return errors.DispatchException(ctx, 502, str(ex)).response()



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import logging
import os

from inspect import isawaitable
from time import time

from functools import partial
from signal import SIG_IGN, SIGINT, SIGTERM
from signal import signal as signal_func

from .protocol import HttpProtocol

from fdk import constants


class Signal(object):
    stopped = False


logger = logging.getLogger(__name__)


def update_current_time(loop):
    """Cache the current time, since it is needed at the end of every
    keep-alive request to update the request timeout time

    :param loop:
    :return:
    """
    global current_time
    current_time = time()
    loop.call_later(1, partial(update_current_time, loop))


def trigger_events(events, loop):
    """Trigger event callbacks (functions or async)

    :param events: one or more sync or async functions to execute
    :param loop: event loop
    """
    for event in events:
        result = event(loop)
        if isawaitable(result):
            loop.run_until_complete(result)


def serve(
    request_handler,
    error_handler,
    debug=False,
    request_timeout=60,
    response_timeout=60,
    keep_alive_timeout=75,
    ssl=None,
    sock=None,
    request_max_size=100000000,
    reuse_port=False,
    loop=None,
    protocol=HttpProtocol,
    backlog=100,
    register_sys_signals=True,
    run_multiple=False,
    run_async=False,
    connections=None,
    signal=Signal(),
    request_class=None,
    access_log=True,
    keep_alive=True,
    is_request_stream=False,
    router=None,
    websocket_max_size=None,
    websocket_max_queue=None,
    websocket_read_limit=2 ** 16,
    websocket_write_limit=2 ** 16,
    state=None,
):
    """Start asynchronous HTTP Server on an individual process.

    :param request_handler: Sanic request handler with middleware
    :param error_handler: Sanic error handler with middleware
    :param debug: enables debug output (slows server)
    :param request_timeout: time in seconds
    :param response_timeout: time in seconds
    :param keep_alive_timeout: time in seconds
    :param ssl: SSLContext
    :param sock: Socket for the server to accept connections from
    :param request_max_size: size in bytes, `None` for no limit
    :param reuse_port: `True` for multiple workers
    :param loop: asyncio compatible event loop
    :param protocol: subclass of asyncio protocol class
    :param request_class: Request class to use
    :param access_log: disable/enable access log
    :param websocket_max_size: enforces the maximum size for
                               incoming messages in bytes.
    :param websocket_max_queue: sets the maximum length of the queue
                                that holds incoming messages.
    :param websocket_read_limit: sets the high-water limit of the buffer for
                                 incoming bytes, the low-water limit is half
                                 the high-water limit.
    :param websocket_write_limit: sets the high-water limit of the buffer for
                                  outgoing bytes, the low-water limit is a
                                  quarter of the high-water limit.
    :param is_request_stream: disable/enable Request.stream
    :return: Nothing
    """
    if not run_async:
        # create new event_loop after fork
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    if debug:
        loop.set_debug(debug)

    connections = connections if connections is not None else set()
    server = partial(
        protocol,
        loop=loop,
        connections=connections,
        signal=signal,
        request_handler=request_handler,
        error_handler=error_handler,
        request_timeout=request_timeout,
        response_timeout=response_timeout,
        keep_alive_timeout=keep_alive_timeout,
        request_max_size=request_max_size,
        request_class=request_class,
        access_log=access_log,
        keep_alive=keep_alive,
        is_request_stream=is_request_stream,
        router=router,
        websocket_max_size=websocket_max_size,
        websocket_max_queue=websocket_max_queue,
        websocket_read_limit=websocket_read_limit,
        websocket_write_limit=websocket_write_limit,
        state=state,
        debug=debug,
    )

    create_server_kwargs = dict(
        ssl=ssl,
        reuse_port=reuse_port,
        sock=sock,
        backlog=backlog,
    )

    if constants.is_py37():
        create_server_kwargs.update(
            start_serving=False
        )

    server_coroutine = loop.create_server(
        server, **create_server_kwargs
    )

    # Instead of pulling time at the end of every request,
    # pull it once per minute
    loop.call_soon(partial(update_current_time, loop))

    if run_async:
        return server_coroutine

    try:
        http_server = loop.run_until_complete(server_coroutine)
    except BaseException:
        logger.exception("Unable to start server")
        return

    # Ignore SIGINT when run_multiple
    if run_multiple:
        signal_func(SIGINT, SIG_IGN)

    # Register signals for graceful termination
    if register_sys_signals:
        _signals = (SIGTERM,) if run_multiple else (SIGINT, SIGTERM)
        for _signal in _signals:
            try:
                loop.add_signal_handler(_signal, loop.stop)
            except NotImplementedError:
                logger.warning(
                    "Sanic tried to use loop.add_signal_handler "
                    "but it is not implemented on this platform."
                )

    def start_serving():
        if constants.is_py37():
            loop.run_until_complete(http_server.start_serving())

    def start():
        pid = os.getpid()
        try:
            logger.debug("Starting worker [%s]", pid)
            if constants.is_py37():
                loop.run_until_complete(http_server.serve_forever())
            else:
                loop.run_forever()
        finally:
            http_server.close()
            loop.run_until_complete(http_server.wait_closed())
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    return start_serving, start



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fdk import context
from fdk import constants
from typing import Union


class Response(object):

    def __init__(self, ctx: context.InvokeContext,
                 response_data: Union[str, bytes] = None,
                 headers: dict = None,
                 status_code: int = 200,
                 response_encoding: str = "utf-8"):
        """
        Creates an FDK-readable response object
        :param ctx: invoke context
        :type ctx: fdk.context.InvokeContext
        :param response_data: function's response data
        :type response_data: str
        :param headers: response headers
        :type headers: dict
        :param status_code: response code
        :type status_code: int
        :param response_encoding: response encoding for strings ("utf-8")
        :type response_encoding: str
        """
        self.ctx = ctx
        self.status_code = status_code
        self.response_data = response_data if response_data else ""
        self.response_encoding = response_encoding

        if headers is None:
            headers = {}
        headers.update({constants.FN_FDK_VERSION:
                        constants.VERSION_HEADER_VALUE,
                        constants.FN_FDK_RUNTIME:
                        constants.RUNTIME_HEADER_VALUE, })
        ctx.SetResponseHeaders(headers, status_code)
        self.ctx = ctx

    def status(self):
        return self.status_code

    def body(self):
        return self.response_data

    def body_bytes(self):
        if isinstance(self.response_data, bytes) or \
                isinstance(self.response_data, bytearray):
            return self.response_data
        else:
            return str(self.response_data).encode(self.response_encoding)

    def context(self):
        return self.ctx



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from asyncio import CancelledError
from traceback import format_exc

from .exceptions import AsyncHTTPException
from .error_handler import ErrorHandler
from .request import Request
from .response import HTTPResponse, StreamingHTTPResponse
from .server import serve

logger = logging.getLogger(__name__)


class AsyncHTTPServer(object):
    def __init__(
        self,
        name=None,
        router=None,
    ):

        self.name = name
        self.router = router
        self.request_class = Request
        self.error_handler = ErrorHandler()
        self.config = {}
        self.debug = None
        self.sock = None
        self.is_running = False
        self.is_request_stream = False
        self.websocket_enabled = False
        self.websocket_tasks = set()

        # Register alternative method names
        self.go_fast = self.run

    async def handle_request(self, request, write_callback, stream_callback):
        """Take a request from the HTTP Server and return a response object
        to be sent back The HTTP Server only expects a response object, so
        exception handling must be done here
        :param request: HTTP Request object
        :param write_callback: Synchronous response function to be
            called with the response as the only argument
        :param stream_callback: Coroutine that handles streaming a
            StreamingHTTPResponse if produced by the handler.
        :return: Nothing
        """
        # Define `response` var here to remove warnings about
        # allocation before assignment below.
        cancelled = False
        try:
            request.app = self
            # -------------------------------------------- #
            # Execute Handler
            # -------------------------------------------- #

            # Fetch handler from router
            handler, uri = self.router.get(
                request.path, request.method)

            request.uri_template = uri
            response = handler(request)
            logger.debug("got response from function")
            res = await response
            body = res.body
            headers = res.headers
            status = res.status
            response = HTTPResponse(
                body_bytes=body, status=status, headers=headers,
            )
        except CancelledError:
            response = None
            cancelled = True
        except Exception as e:
            if isinstance(e, AsyncHTTPException):
                response = self.error_handler.default(
                    request=request, exception=e
                )
            elif self.debug:
                response = HTTPResponse(
                    "Error while handling error: {}\nStack: {}".format(
                        e, format_exc()
                    ),
                    status=502,
                )
            else:
                response = HTTPResponse(
                    "An error occurred while handling an error", status=502
                )
        finally:
            if cancelled:
                raise CancelledError()

        if isinstance(response, StreamingHTTPResponse):
            await stream_callback(response)
        else:
            write_callback(response)

    def run(self, sock=None, loop=None):
        return serve(
            self.handle_request, ErrorHandler(),
            sock=sock, loop=loop
        )



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime as dt
import io
import os
import random
from fdk import constants
from fdk import headers as hs
from fdk import log
from collections import namedtuple


class InvokeContext(object):

    def __init__(self, app_id, app_name, fn_id, fn_name, call_id,
                 content_type="application/octet-stream",
                 deadline=None, config=None,
                 headers=None, request_url=None,
                 method="POST", fn_format=None,
                 tracing_context=None):
        """
        Request context here to be a placeholder
        for request-specific attributes
        :param app_id: Fn App ID
        :type app_id: str
        :param app_name: Fn App name
        :type app_name: str
        :param fn_id: Fn App Fn ID
        :type fn_id: str
        :param fn_name: Fn name
        :type fn_name: str
        :param call_id: Fn call ID
        :type call_id: str
        :param content_type: request content type
        :type content_type: str
        :param deadline: request deadline
        :type deadline: str
        :param config: an app/fn config
        :type config: dict
        :param headers: request headers
        :type headers: dict
        :param request_url: request URL
        :type request_url: str
        :param method: request method
        :type method: str
        :param fn_format: function format
        :type fn_format: str
        :param tracing_context: tracing context
        :type tracing_context: TracingContext
        """
        self.__app_id = app_id
        self.__fn_id = fn_id
        self.__call_id = call_id
        self.__config = config if config else {}
        self.__headers = headers if headers else {}
        self.__http_headers = {}
        self.__deadline = deadline
        self.__content_type = content_type
        self._request_url = request_url
        self._method = method
        self.__response_headers = {}
        self.__fn_format = fn_format
        self.__app_name = app_name
        self.__fn_name = fn_name
        self.__tracing_context = tracing_context if tracing_context else None

        log.log("request headers. gateway: {0} {1}"
                .format(self.__is_gateway(), headers))

        if self.__is_gateway():
            self.__headers = hs.decap_headers(headers, True)
            self.__http_headers = hs.decap_headers(headers, False)

    def AppID(self):
        return self.__app_id

    def AppName(self):
        return self.__app_name

    def FnID(self):
        return self.__fn_id

    def FnName(self):
        return self.__fn_name

    def CallID(self):
        return self.__call_id

    def Config(self):
        return self.__config

    def Headers(self):
        return self.__headers

    def HTTPHeaders(self):
        return self.__http_headers

    def Format(self):
        return self.__fn_format

    def TracingContext(self):
        return self.__tracing_context

    def Deadline(self):
        if self.__deadline is None:
            now = dt.datetime.now(dt.timezone.utc).astimezone()
            now += dt.timedelta(0, float(constants.DEFAULT_DEADLINE))
            return now.isoformat()
        return self.__deadline

    def SetResponseHeaders(self, headers, status_code):
        log.log("setting headers. gateway: {0}".format(self.__is_gateway()))
        if self.__is_gateway():
            headers = hs.encap_headers(headers, status=status_code)

        for k, v in headers.items():
            self.__response_headers[k.lower()] = v

    def GetResponseHeaders(self):
        return self.__response_headers

    def RequestURL(self):
        return self._request_url

    def Method(self):
        return self._method

    def __is_gateway(self):
        return (constants.FN_INTENT in self.__headers
                and self.__headers.get(constants.FN_INTENT)
                == constants.INTENT_HTTP_REQUEST)


class TracingContext(object):

    def __init__(self, is_tracing_enabled, trace_collector_url,
                 trace_id, span_id, parent_span_id,
                 is_sampled, flags):
        """
        Tracing context here to be a placeholder
        for tracing-specific attributes
        :param is_tracing_enabled: tracing enabled flag
        :type is_tracing_enabled: bool
        :param trace_collector_url: APM Trace Collector Endpoint URL
        :type trace_collector_url: str
        :param trace_id: Trace ID
        :type trace_id: str
        :param span_id: Span ID
        :type span_id: str
        :param parent_span_id: Parent Span ID
        :type parent_span_id: str
        :param is_sampled: Boolean for emmitting spans
        :type is_sampled: int (0 or 1)
        :param flags: Debug flags
        :type flags: int (0 or 1)
        """
        self.__is_tracing_enabled = is_tracing_enabled
        self.__trace_collector_url = trace_collector_url
        self.__trace_id = trace_id
        self.__span_id = span_id
        self.__parent_span_id = parent_span_id
        self.__is_sampled = is_sampled
        self.__flags = flags
        self.__app_name = os.environ.get(constants.FN_APP_NAME)
        self.__app_id = os.environ.get(constants.FN_APP_ID)
        self.__fn_name = os.environ.get(constants.FN_NAME)
        self.__fn_id = os.environ.get(constants.FN_ID)

        self.__zipkin_attrs = self.__create_zipkin_attrs(is_tracing_enabled)

    def is_tracing_enabled(self):
        return self.__is_tracing_enabled

    def trace_collector_url(self):
        return self.__trace_collector_url

    def trace_id(self):
        return self.__trace_id

    def span_id(self):
        return self.__span_id

    def parent_span_id(self):
        return self.__parent_span_id

    def is_sampled(self):
        return bool(self.__is_sampled)

    def flags(self):
        return self.__flags

    def zipkin_attrs(self):
        return self.__zipkin_attrs

    # this is a helper method specific for py_zipkin
    def __create_zipkin_attrs(self, is_tracing_enabled):
        ZipkinAttrs = namedtuple(
            "ZipkinAttrs",
            "trace_id, span_id, parent_span_id, is_sampled, flags"
        )

        trace_id = self.__trace_id
        span_id = self.__span_id
        parent_span_id = self.__parent_span_id
        is_sampled = bool(self.__is_sampled)
        trace_flags = self.__flags

        # As the fnLb sends the parent_span_id as the span_id
        # assign the parent span id as the span id.
        if is_tracing_enabled:
            parent_span_id = span_id
            span_id = generate_id()

        zipkin_attrs = ZipkinAttrs(
            trace_id,
            span_id,
            parent_span_id,
            is_sampled,
            trace_flags
        )
        return zipkin_attrs

    def service_name(self, override=None):
        # in case of missing app and function name env variables
        service_name = (
            override
            if override is not None
            else str(self.__app_name) + "::" + str(self.__fn_name)
        )
        return service_name.lower()

    def annotations(self):
        annotations = {
            "generatedBy": "faas",
            "appName": self.__app_name,
            "appID": self.__app_id,
            "fnName": self.__fn_name,
            "fnID": self.__fn_id,
        }
        return annotations


def generate_id():
    return "{:016x}".format(random.getrandbits(64))


def context_from_format(format_def: str, **kwargs) -> (
        InvokeContext, io.BytesIO):
    """
    Creates a context from request
    :param format_def: function format
    :type format_def: str
    :param kwargs: request-specific map of parameters
    :return: invoke context and data
    :rtype: tuple
    """

    app_id = os.environ.get(constants.FN_APP_ID)
    fn_id = os.environ.get(constants.FN_ID)
    app_name = os.environ.get(constants.FN_APP_NAME)
    fn_name = os.environ.get(constants.FN_NAME)
    # the tracing enabled env variable is passed as a "0" or "1" string
    # and therefore needs to be converted appropriately.
    is_tracing_enabled = os.environ.get(constants.OCI_TRACING_ENABLED)
    is_tracing_enabled = (
        bool(int(is_tracing_enabled))
        if is_tracing_enabled is not None
        else False
    )
    trace_collector_url = os.environ.get(constants.OCI_TRACE_COLLECTOR_URL)

    if format_def == constants.HTTPSTREAM:
        data = kwargs.get("data")
        headers = kwargs.get("headers")

        # zipkin tracing http headers
        trace_id = span_id = parent_span_id = is_sampled = trace_flags = None
        tracing_context = None
        if is_tracing_enabled:
            # we generate the trace_id if tracing is enabled
            # but the traceId zipkin header is missing.
            trace_id = headers.get(constants.X_B3_TRACEID)
            trace_id = generate_id() if trace_id is None else trace_id

            span_id = headers.get(constants.X_B3_SPANID)
            parent_span_id = headers.get(constants.X_B3_PARENTSPANID)

            # span_id is also generated if the zipkin header is missing.
            span_id = generate_id() if span_id is None else span_id

            # is_sampled should be a boolean in the form of a "0/1" but
            # legacy samples have them as "False/True"
            is_sampled = headers.get(constants.X_B3_SAMPLED)
            is_sampled = int(is_sampled) if is_sampled is not None else 1

            # not currently used but is defined by the zipkin headers standard
            trace_flags = headers.get(constants.X_B3_FLAGS)

        # tracing context will be an empty object
        # if tracing is not enabled or the flag is missing.
        # this prevents the customer code from failing if they decide to
        # disable tracing. An empty tracing context will not
        # emit spans due to is_sampled being None.
        tracing_context = TracingContext(
            is_tracing_enabled,
            trace_collector_url,
            trace_id,
            span_id,
            parent_span_id,
            is_sampled,
            trace_flags
        )

        method = headers.get(constants.FN_HTTP_METHOD)
        request_url = headers.get(constants.FN_HTTP_REQUEST_URL)
        deadline = headers.get(constants.FN_DEADLINE)
        call_id = headers.get(constants.FN_CALL_ID)
        content_type = headers.get(constants.CONTENT_TYPE)

        ctx = InvokeContext(
            app_id, app_name, fn_id, fn_name, call_id,
            content_type=content_type,
            deadline=deadline,
            config=os.environ,
            headers=headers,
            method=method,
            request_url=request_url,
            fn_format=constants.HTTPSTREAM,
            tracing_context=tracing_context,
        )

        return ctx, data



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fdk.async_http import protocol


def http_protocol(**kwargs):
    return protocol.HttpProtocol(
        loop=None, request_handler=None, error_handler=None,
        **kwargs)


def test_keep_alive_time_left_before_time_is_set():
    p = http_protocol()

    protocol.current_time = None
    p._last_request_time = None

    time_left = p.keep_alive_time_left()
    assert 0 == time_left


def test_keep_alive_time_left_after_time_is_set():
    p = http_protocol(keep_alive_timeout=15)

    protocol.current_time = 96
    p._last_response_time = 64

    time_left = p.keep_alive_time_left()
    assert 15 - 32 == time_left



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from fdk import constants


def get_delayed_module_init_class():
    if constants.is_py37():
        return Python37DelayedImport
    else:
        return Python35plusDelayedImport


class PythonDelayedImportAbstraction(object):

    def __init__(self, func_module_path):
        self._mod_path = func_module_path
        self._executed = False

    @property
    def executed(self):
        return self._executed

    @executed.setter
    def executed(self, exec_flag):
        self._executed = exec_flag

    def get_module(self):
        raise Exception("Not implemented")


class Python35plusDelayedImport(PythonDelayedImportAbstraction):

    def __init__(self, func_module_path):
        self._func_module = None
        super(Python35plusDelayedImport, self).__init__(func_module_path)

    def get_module(self):
        if not self.executed:
            from importlib.machinery import SourceFileLoader
            fname, ext = os.path.splitext(
                os.path.basename(self._mod_path))
            self._func_module = SourceFileLoader(fname, self._mod_path)\
                .load_module()
            self.executed = True

        return self._func_module


class Python37DelayedImport(PythonDelayedImportAbstraction):

    def import_from_source(self):
        from importlib import util
        func_module_spec = util.spec_from_file_location(
            "func", self._mod_path
        )
        func_module = util.module_from_spec(func_module_spec)
        self._func_module_spec = func_module_spec
        self._func_module = func_module

    def get_module(self):
        if not self.executed:
            self.import_from_source()
            self._func_module_spec.loader.exec_module(
                self._func_module)
            self.executed = True

        return self._func_module


class Function(object):

    def __init__(self, func_module_path, entrypoint="handler"):
        dm = get_delayed_module_init_class()
        self._delayed_module_class = dm(func_module_path)
        self._entrypoint = entrypoint

    def handler(self):
        mod = self._delayed_module_class.get_module()
        return getattr(mod, self._entrypoint)



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Defines basics of HTTP standard."""

# TODO: use native python tools to build this map at the runtime
STATUS_CODES = {
    100: b"Continue",
    101: b"Switching Protocols",
    102: b"Processing",
    200: b"OK",
    201: b"Created",
    202: b"Accepted",
    203: b"Non-Authoritative Information",
    204: b"No Content",
    205: b"Reset Content",
    206: b"Partial Content",
    207: b"Multi-Status",
    208: b"Already Reported",
    226: b"IM Used",
    300: b"Multiple Choices",
    301: b"Moved Permanently",
    302: b"Found",
    303: b"See Other",
    304: b"Not Modified",
    305: b"Use Proxy",
    307: b"Temporary Redirect",
    308: b"Permanent Redirect",
    400: b"Bad Request",
    401: b"Unauthorized",
    402: b"Payment Required",
    403: b"Forbidden",
    404: b"Not Found",
    405: b"Method Not Allowed",
    406: b"Not Acceptable",
    407: b"Proxy Authentication Required",
    408: b"Request Timeout",
    409: b"Conflict",
    410: b"Gone",
    411: b"Length Required",
    412: b"Precondition Failed",
    413: b"Request Entity Too Large",
    414: b"Request-URI Too Long",
    415: b"Unsupported Media Type",
    416: b"Requested Range Not Satisfiable",
    417: b"Expectation Failed",
    418: b"I'm a teapot",
    422: b"Unprocessable Entity",
    423: b"Locked",
    424: b"Failed Dependency",
    426: b"Upgrade Required",
    428: b"Precondition Required",
    429: b"Too Many Requests",
    431: b"Request Header Fields Too Large",
    451: b"Unavailable For Legal Reasons",
    500: b"Internal Server Error",
    501: b"Not Implemented",
    502: b"Bad Gateway",
    503: b"Service Unavailable",
    504: b"Gateway Timeout",
    505: b"HTTP Version Not Supported",
    506: b"Variant Also Negotiates",
    507: b"Insufficient Storage",
    508: b"Loop Detected",
    510: b"Not Extended",
    511: b"Network Authentication Required",
}

# According to https://tools.ietf.org/html/rfc2616#section-7.1
_ENTITY_HEADERS = frozenset(
    [
        "allow",
        "content-encoding",
        "content-language",
        "content-length",
        "content-location",
        "content-md5",
        "content-range",
        "content-type",
        "expires",
        "last-modified",
        "extension-header",
    ]
)

# According to https://tools.ietf.org/html/rfc2616#section-13.5.1
_HOP_BY_HOP_HEADERS = frozenset(
    [
        "connection",
        "keep-alive",
        "proxy-authenticate",
        "proxy-authorization",
        "te",
        "trailers",
        "transfer-encoding",
        "upgrade",
    ]
)


def has_message_body(status):
    """
    According to the following RFC message body and length SHOULD NOT
    be included in responses status 1XX, 204 and 304.
    https://tools.ietf.org/html/rfc2616#section-4.4
    https://tools.ietf.org/html/rfc2616#section-4.3
    """
    return status not in (204, 304) and not (100 <= status < 200)


def is_entity_header(header):
    """Checks if the given header is an Entity Header"""
    return header.lower() in _ENTITY_HEADERS


def is_hop_by_hop_header(header):
    """Checks if the given header is a Hop By Hop header"""
    return header.lower() in _HOP_BY_HOP_HEADERS


def remove_entity_headers(headers, allowed=("content-location", "expires")):
    """
    Removes all the entity headers present in the headers given.
    According to RFC 2616 Section 10.3.5,
    Content-Location and Expires are allowed as for the
    "strong cache validator".
    https://tools.ietf.org/html/rfc2616#section-10.3.5

    returns the headers without the entity headers
    """
    allowed = set([h.lower() for h in allowed])
    headers = {
        header: value
        for header, value in headers.items()
        if not is_entity_header(header) or header.lower() in allowed
    }
    return headers


TRACEBACK_STYLE = """
    <style>
        body {
            padding: 20px;
            font-family: Arial, sans-serif;
        }

        p {
            margin: 0;
        }

        .summary {
            padding: 10px;
        }

        h1 {
            margin-bottom: 0;
        }

        h3 {
            margin-top: 10px;
        }

        h3 code {
            font-size: 24px;
        }

        .frame-line > * {
            padding: 5px 10px;
        }

        .frame-line {
            margin-bottom: 5px;
        }

        .frame-code {
            font-size: 16px;
            padding-left: 30px;
        }

        .tb-wrapper {
            border: 1px solid #f3f3f3;
        }

        .tb-header {
            background-color: #f3f3f3;
            padding: 5px 10px;
        }

        .tb-border {
            padding-top: 20px;
        }

        .frame-descriptor {
            background-color: #e2eafb;
        }

        .frame-descriptor {
            font-size: 14px;
        }
    </style>
"""

TRACEBACK_WRAPPER_HTML = """
    <html>
        <head>
            {style}
        </head>
        <body>
            {inner_html}
            <div class="summary">
                <p>
                <b>{exc_name}: {exc_value}</b>
                    while handling path <code>{path}</code>
                </p>
            </div>
        </body>
    </html>
"""

TRACEBACK_WRAPPER_INNER_HTML = """
    <h1>{exc_name}</h1>
    <h3><code>{exc_value}</code></h3>
    <div class="tb-wrapper">
        <p class="tb-header">Traceback (most recent call last):</p>
        {frame_html}
    </div>
"""

TRACEBACK_BORDER = """
    <div class="tb-border">
        <b><i>
            The above exception was the direct cause of the
            following exception:
        </i></b>
    </div>
"""

TRACEBACK_LINE_HTML = """
    <div class="frame-line">
        <p class="frame-descriptor">
            File {0.filename}, line <i>{0.lineno}</i>,
            in <code><b>{0.name}</b></code>
        </p>
        <p class="frame-code"><code>{0.line}</code></p>
    </div>
"""

INTERNAL_SERVER_ERROR_HTML = """
    <h1>Internal Server Error</h1>
    <p>
        The server encountered an internal error and cannot complete
        your request.
    </p>
"""


_excs = {}


def add_status_code(code):
    """
    Decorator used for adding exceptions to _sanic_exceptions.
    """

    def class_decorator(cls):
        cls.status_code = code
        _excs[code] = cls
        return cls

    return class_decorator


class AsyncHTTPException(Exception):
    def __init__(self, message, status_code=None):
        super().__init__(message)

        if status_code is not None:
            self.status_code = status_code


@add_status_code(404)
class NotFound(AsyncHTTPException):
    pass


@add_status_code(400)
class InvalidUsage(AsyncHTTPException):
    pass


@add_status_code(405)
class MethodNotSupported(AsyncHTTPException):
    def __init__(self, message, method, allowed_methods):
        super().__init__(message)
        self.headers = dict()
        self.headers["Allow"] = ", ".join(allowed_methods)
        if method in ["HEAD", "PATCH", "PUT", "DELETE"]:
            self.headers["Content-Length"] = 0


@add_status_code(500)
class ServerError(AsyncHTTPException):
    pass


@add_status_code(503)
class ServiceUnavailable(AsyncHTTPException):
    """The server is currently unavailable (because it is overloaded or
    down for maintenance). Generally, this is a temporary state."""

    pass


class URLBuildError(ServerError):
    pass


class FileNotFound(NotFound):
    def __init__(self, message, path, relative_url):
        super().__init__(message)
        self.path = path
        self.relative_url = relative_url


@add_status_code(408)
class RequestTimeout(AsyncHTTPException):
    """The Web server (running the Web site) thinks that there has been too
    long an interval of time between 1) the establishment of an IP
    connection (socket) between the client and the server and
    2) the receipt of any data on that socket, so the server has dropped
    the connection. The socket connection has actually been lost - the Web
    server has 'timed out' on that particular socket connection.
    """

    pass


@add_status_code(413)
class PayloadTooLarge(AsyncHTTPException):
    pass


class HeaderNotFound(InvalidUsage):
    pass


@add_status_code(416)
class ContentRangeError(AsyncHTTPException):
    def __init__(self, message, content_range):
        super().__init__(message)
        self.headers = {
            "Content-Type": "text/plain",
            "Content-Range": "bytes */%s" % (content_range.total,),
        }


@add_status_code(403)
class Forbidden(AsyncHTTPException):
    pass


class InvalidRangeType(ContentRangeError):
    pass


class PyFileError(Exception):
    def __init__(self, file):
        super().__init__("could not execute config file %s", file)


@add_status_code(401)
class Unauthorized(AsyncHTTPException):
    """
    Unauthorized exception (401 HTTP status code).

    :param message: Message describing the exception.
    :param status_code: HTTP Status code.
    :param scheme: Name of the authentication scheme to be used.

    When present, kwargs is used to complete the WWW-Authentication header.

    Examples::

        # With a Basic auth-scheme, realm MUST be present:
        raise Unauthorized("Auth required.",
                           scheme="Basic",
                           realm="Restricted Area")

        # With a Digest auth-scheme, things are a bit more complicated:
        raise Unauthorized("Auth required.",
                           scheme="Digest",
                           realm="Restricted Area",
                           qop="auth, auth-int",
                           algorithm="MD5",
                           nonce="abcdef",
                           opaque="zyxwvu")

        # With a Bearer auth-scheme, realm is optional so you can write:
        raise Unauthorized("Auth required.", scheme="Bearer")

        # or, if you want to specify the realm:
        raise Unauthorized("Auth required.",
                           scheme="Bearer",
                           realm="Restricted Area")
    """

    def __init__(self, message, status_code=None, scheme=None, **kwargs):
        super().__init__(message, status_code)

        # if auth-scheme is specified, set "WWW-Authenticate" header
        if scheme is not None:
            values = ['{!s}="{!s}"'.format(k, v) for k, v in kwargs.items()]
            challenge = ", ".join(values)

            self.headers = {
                "WWW-Authenticate": "{} {}".format(scheme, challenge).rstrip()
            }


def abort(status_code, message=None):
    """
    Raise an exception based on AsyncHTTPException. Returns the HTTP response
    message appropriate for the given status code, unless provided.

    :param status_code: The HTTP status code to return.
    :param message: The HTTP response body. Defaults to the messages
                    in response.py for the given status code.
    """
    if message is None:
        message = STATUS_CODES.get(status_code)
        # These are stored as bytes in the STATUS_CODES dict
        message = message.decode("utf8")
    exc = _excs.get(status_code, AsyncHTTPException)
    raise exc(message=message, status_code=status_code)



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import socket

from fdk import customer_code
from fdk import event_handler
from fdk import log

from fdk.async_http import app
from fdk.async_http import router


def handle(handle_code: customer_code.Function, port: int = 5000):
    """
    FDK entry point
    :param handle_code: customer's code
    :type handle_code: fdk.customer_code.Function
    :param port: TCP port to start an FDK at
    :type port: int
    :return: None
    """
    host = "localhost"
    log.log("entering handle")

    log.log("Starting HTTP server on "
            "TCP socket: {0}:{1}".format(host, port))
    loop = asyncio.get_event_loop()

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("localhost", port))

    rtr = router.Router()
    rtr.add("/call", frozenset({"POST"}),
            event_handler.event_handle(handle_code))
    srv = app.AsyncHTTPServer(name="fdk-tcp-debug", router=rtr)
    start_serving, server_forever = srv.run(sock=sock, loop=loop)
    start_serving()
    server_forever()



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import json

from fdk import response

xml = """<!DOCTYPE mensaje SYSTEM "record.dtd">
<record>
    <player_birthday>1979-09-23</player_birthday>
    <player_name>Orene Ai'i</player_name>
    <player_team>Blues</player_team>
    <player_id>453</player_id>
    <player_height>170</player_height>
    <player_position>FW</player_position>
    <player_weight>75</player_weight>
</record>"""


def code404(ctx, **kwargs):
    return response.Response(ctx, status_code=404)


def code502(ctx, **kwargs):
    return response.Response(ctx, status_code=502)


def code504(ctx, **kwargs):
    return response.Response(ctx, status_code=504)


def dummy_func(ctx, data=None):
    if data is not None and len(data) > 0:
        body = json.loads(data)
    else:
        body = {"name": "World"}
    return "Hello {0}".format(body.get("name"))


def encaped_header(ctx, **kwargs):
    return response.Response(
        ctx, response_data="OK", status_code=200,
        headers=ctx.Headers())


def content_type(ctx, data=None):
    return response.Response(
        ctx, response_data="OK", status_code=200,
        headers={"Content-Type": "application/json"})


def custom_response(ctx, data=None):
    return response.Response(
        ctx,
        response_data=dummy_func(ctx, data=data),
        status_code=201)


def expectioner(ctx, data=None):
    raise Exception("custom_error")


def none_func(ctx, data=None):
    return


def timed_sleepr(timeout):
    def sleeper(ctx, data=None):
        time.sleep(timeout)

    return sleeper


async def coro(ctx, **kwargs):
    return "hello from coro"


def valid_xml(ctx, **kwargs):
    return response.Response(
        ctx, response_data=xml, headers={
            "content-type": "application/xml",
        }
    )


def invalid_xml(ctx, **kwargs):
    return response.Response(
        ctx, response_data=json.dumps(xml), headers={
            "content-type": "application/xml",
        }
    )


def verify_request_headers(ctx, **kwargs):
    return response.Response(
        ctx,
        response_data=json.dumps(xml),
        headers=ctx.Headers()
    )


def access_request_url(ctx, **kwargs):
    method = ctx.Method()
    request_url = ctx.RequestURL()
    return response.Response(
        ctx, response_data="OK", headers={
            "Response-Request-URL": request_url,
            "Request-Method": method,
        }
    )


captured_context = None


def setup_context_capture():
    global captured_context
    captured_context = None


def get_captured_context():
    global captured_context
    my_context = captured_context
    captured_context = None
    return my_context


def capture_request_ctx(ctx, **kwargs):
    global captured_context
    captured_context = ctx
    return response.Response(ctx, response_data="OK")


def binary_result(ctx, **kwargs):
    return response.Response(ctx, response_data=bytes([1, 2, 3, 4, 5]))



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from fdk import constants
from fdk import customer_code

from fdk.tests import funcs


@pytest.mark.skipif(constants.is_py37(),
                    reason="this test is for Python 3.5.2+ only")
def test_py352plus():
    dm = customer_code.Python35plusDelayedImport(funcs.__file__)
    assert dm.executed is False

    m = dm.get_module()
    assert dm.executed is True
    assert m is not None


@pytest.mark.skipif(not constants.is_py37(),
                    reason="this test is for Python 3.7.1+ only")
def test_py37():
    dm = customer_code.Python37DelayedImport(funcs.__file__)
    assert dm.executed is False

    m = dm.get_module()
    assert dm.executed is True
    assert m is not None


def test_generic_delayed_loader():
    f = customer_code.Function(
        funcs.__file__, entrypoint="content_type")
    assert f is not None
    assert f._delayed_module_class.executed is False

    h = f.handler()
    assert h is not None
    assert f._delayed_module_class.executed is True



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from fdk import constants
from fdk.context import TracingContext
from fdk.context import InvokeContext
from fdk.context import context_from_format

app_id = "some_app_id"
app_name = "mock_app"
fn_id = "some_fn_id"
fn_name = "mock_fn"

is_tracing_enabled = True
trace_collector_url = "some_url_endpoint"
trace_id = "some_trace_id"
span_id = "some_span_id"
parent_span_id = "some_parent_span_id"
is_sampled = 1
trace_flags = 0

headers = {
    'host': 'localhost',
    'user-agent': 'curl/7.64.1',
    'content-length': '0',
    'accept': '*/*',
    'content-type': 'application/json',
    'fn-call-id': '01EY6NW519NG8G00GZJ000001Z',
    'fn-deadline': '2021-02-10T19:15:19Z',
    'accept-encoding': 'gzip',
    'x-b3-traceid': trace_id,
    'x-b3-spanid': span_id,
    'x-b3-parentspanid': parent_span_id,
    'x-b3-sampled': is_sampled,
    'x-b3-flags': trace_flags
}
call_id = headers.get(constants.FN_CALL_ID)
content_type = headers.get(constants.CONTENT_TYPE)
deadline = headers.get(constants.FN_DEADLINE)
method = headers.get(constants.FN_HTTP_METHOD)
request_url = headers.get(constants.FN_HTTP_REQUEST_URL)


def tracing_context():
    return TracingContext(
        is_tracing_enabled,
        trace_collector_url,
        trace_id,
        span_id,
        parent_span_id,
        is_sampled,
        trace_flags
    )


def invoke_context():
    return InvokeContext(
        app_id, app_name, fn_id, fn_name, call_id,
        content_type=content_type,
        deadline=deadline,
        config=os.environ,
        headers=headers,
        method=method,
        request_url=request_url,
        fn_format=constants.HTTPSTREAM,
        tracing_context=tracing_context(),
    )


def test_context_from_format_returns_correct_objects_when_tracing_enabled():
    os.environ[constants.FN_APP_ID] = app_id
    os.environ[constants.FN_ID] = fn_id
    os.environ[constants.FN_APP_NAME] = app_name
    os.environ[constants.FN_NAME] = fn_name
    os.environ[constants.OCI_TRACE_COLLECTOR_URL] = trace_collector_url
    os.environ[constants.OCI_TRACING_ENABLED] = "1"

    mock_invoke_context = invoke_context()
    mock_data = "some_data"

    actual_invoke_context, data = context_from_format(
        constants.HTTPSTREAM,
        headers=headers,
        data=mock_data
    )

    actual = actual_invoke_context.TracingContext().__dict__
    expected = mock_invoke_context.TracingContext().__dict__

    attrs = "_TracingContext__zipkin_attrs"
    filtered_actual = {
        key: value for (key, value) in actual.items() if key != attrs
    }
    filtered_expected = {
        key: value for (key, value) in expected.items() if key != attrs
    }

    assert filtered_actual == filtered_expected
    assert actual[attrs].trace_id == expected[attrs].trace_id
    assert actual[attrs].span_id != expected[attrs].span_id
    assert actual[attrs].span_id != span_id
    assert actual[attrs].parent_span_id == expected[attrs].parent_span_id
    assert actual[attrs].is_sampled == expected[attrs].is_sampled
    assert actual[attrs].flags == expected[attrs].flags

    assert data == mock_data

    os.environ.pop(constants.FN_APP_ID)
    os.environ.pop(constants.FN_ID)
    os.environ.pop(constants.FN_APP_NAME)
    os.environ.pop(constants.FN_NAME)
    os.environ.pop(constants.OCI_TRACE_COLLECTOR_URL)
    os.environ.pop(constants.OCI_TRACING_ENABLED)


def test_context_from_format_returns_correct_objects_when_tracing_disabled():
    os.environ[constants.FN_APP_ID] = app_id
    os.environ[constants.FN_ID] = fn_id
    os.environ[constants.FN_APP_NAME] = app_name
    os.environ[constants.FN_NAME] = fn_name
    os.environ[constants.OCI_TRACING_ENABLED] = "0"

    empty_tracing_context = TracingContext(
        False, None, None, None, None, None, None
    )
    mock_data = "some_data"

    actual_invoke_context, data = context_from_format(
        constants.HTTPSTREAM,
        headers=headers,
        data=mock_data
    )

    actual_tracing_context = actual_invoke_context.TracingContext().__dict__
    assert actual_tracing_context == empty_tracing_context.__dict__
    assert data == mock_data

    os.environ.pop(constants.FN_APP_ID)
    os.environ.pop(constants.FN_ID)
    os.environ.pop(constants.FN_APP_NAME)
    os.environ.pop(constants.FN_NAME)
    os.environ.pop(constants.OCI_TRACING_ENABLED)


def test_zipkin_attrs():
    mock_tracing_context = tracing_context()
    zipkin_attrs = mock_tracing_context.zipkin_attrs()

    assert zipkin_attrs.trace_id == trace_id
    assert zipkin_attrs.span_id != span_id
    assert zipkin_attrs.parent_span_id == span_id
    assert zipkin_attrs.is_sampled is True
    assert zipkin_attrs.flags == 0


def test_tracing_context():
    os.environ[constants.FN_APP_NAME] = app_name
    os.environ[constants.FN_NAME] = fn_name
    mock_tracing_context = tracing_context()

    assert mock_tracing_context.trace_collector_url() == trace_collector_url
    assert mock_tracing_context.trace_id() == trace_id
    assert mock_tracing_context.span_id() == span_id
    assert mock_tracing_context.parent_span_id() == parent_span_id
    assert mock_tracing_context.is_sampled() == bool(is_sampled)
    assert mock_tracing_context.flags() == trace_flags
    assert mock_tracing_context.service_name() == "mock_app::mock_fn"

    os.environ.pop(constants.FN_APP_NAME)
    os.environ.pop(constants.FN_NAME)



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from fdk import version

ASYNC_IO_READ_BUFFER = 65536
DEFAULT_DEADLINE = 30

HTTPSTREAM = "http-stream"
INTENT_HTTP_REQUEST = "httprequest"

# env vars
FN_FORMAT = "FN_FORMAT"
FN_LISTENER = "FN_LISTENER"
FN_APP_ID = "FN_APP_ID"
FN_ID = "FN_FN_ID"
FN_LOGFRAME_NAME = "FN_LOGFRAME_NAME"
FN_LOGFRAME_HDR = "FN_LOGFRAME_HDR"
FN_APP_NAME = "FN_APP_NAME"
FN_NAME = "FN_FN_NAME"
OCI_TRACE_COLLECTOR_URL = "OCI_TRACE_COLLECTOR_URL"
OCI_TRACING_ENABLED = "OCI_TRACING_ENABLED"


# headers are lower case TODO(denis): why?
FN_INTENT = "fn-intent"
FN_HTTP_PREFIX = "fn-http-h-"
FN_HTTP_STATUS = "fn-http-status"
FN_DEADLINE = "fn-deadline"
FN_FDK_VERSION = "fn-fdk-version"
FN_FDK_RUNTIME = "fn-fdk-runtime"
FN_HTTP_REQUEST_URL = "fn-http-request-url"
FN_CALL_ID = "fn-call-id"
FN_HTTP_METHOD = "fn-http-method"
CONTENT_TYPE = "content-type"
CONTENT_LENGTH = "content-length"
X_B3_TRACEID = "x-b3-traceid"
X_B3_SPANID = "x-b3-spanid"
X_B3_PARENTSPANID = "x-b3-parentspanid"
X_B3_SAMPLED = "x-b3-sampled"
X_B3_FLAGS = "x-b3-flags"

FN_ENFORCED_RESPONSE_CODES = [200, 502, 504]
FN_DEFAULT_RESPONSE_CODE = 200

VERSION_HEADER_VALUE = "fdk-python/" + version.VERSION
RUNTIME_HEADER_VALUE = "python/{}.{}.{} {}".format(
    sys.version_info.major, sys.version_info.minor, sys.version_info.micro,
    sys.version_info.releaselevel)


# todo: python 3.8 is on its way, make more flexible
def is_py37():
    py_version = sys.version_info
    return (py_version.major, py_version.minor) == (3, 7)



#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fdk import constants
from fdk import response


class DispatchException(Exception):

    def __init__(self, ctx, status, message):
        """
        JSON response with error
        :param status: HTTP status code
        :param message: error message
        """
        self.status = status
        self.message = message
        self.ctx = ctx

    def response(self):
        resp_headers = {
            constants.CONTENT_TYPE: "application/json; charset=utf-8",
        }
        return response.Response(
            self.ctx,
            response_data=self.message,
            headers=resp_headers,
            status_code=self.status
        )
